from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.migrations.shared import GraphMigration, MigrationResult
from infrahub.log import get_logger

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class DeleteDuplicateHasValueEdgesQuery(Query):
    name = "delete_duplicate_has_value_edges"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        query = """
// -------------------
// find Attribute nodes with multiple identical edges to AttributeValue nodes with the same value
// -------------------
MATCH (a:Attribute)-[e:HAS_VALUE]->(av:AttributeValue)
WITH a, e.branch AS branch, e.branch_level AS branch_level, e.status AS status, e.from AS from, e.to AS to,
    av.value AS attr_val, av.is_default AS attr_default, COUNT(*) AS num_duplicate_edges
WHERE num_duplicate_edges > 1
// -------------------
// get the the one AttributeValue we want to use
// -------------------
WITH DISTINCT a, branch, branch_level, status, from, to, attr_val, attr_default
WITH attr_val, attr_default, collect([a, branch, branch_level, status, from, to]) AS details_list
CALL {
    WITH attr_val, attr_default
    MATCH (av:AttributeValue {value: attr_val, is_default: attr_default})
    RETURN av AS the_one_av
    ORDER by %(id_func)s(av) ASC
    LIMIT 1
}
UNWIND details_list AS details_item
WITH attr_val, attr_default, the_one_av,
    details_item[0] AS a, details_item[1] AS branch, details_item[2] AS branch_level,
    details_item[3] AS status, details_item[4] AS from, details_item[5] AS to
// -------------------
// get/create the one edge to keep
// -------------------
CREATE (a)-[fresh_e:HAS_VALUE {branch: branch, branch_level: branch_level, status: status, from: from}]->(the_one_av)
SET fresh_e.to = to
WITH a, branch, status, from, to, attr_val, attr_default, %(id_func)s(fresh_e) AS e_id_to_keep
// -------------------
// get the identical edges for a given set of Attribute node, edge properties, AttributeValue.value
// -------------------
CALL {
    // -------------------
    // delete the duplicate edges a given set of Attribute node, edge properties, AttributeValue.value
    // -------------------
    WITH a, branch, status, from, to, attr_val, attr_default, e_id_to_keep
    MATCH (a)-[e:HAS_VALUE]->(av:AttributeValue {value: attr_val, is_default: attr_default})
    WHERE %(id_func)s(e) <> e_id_to_keep
    AND e.branch = branch AND e.status = status AND e.from = from
    AND (e.to = to OR (e.to IS NULL AND to IS NULL))
    DELETE e
}
// -------------------
// delete any orphaned AttributeValue nodes
// -------------------
WITH NULL AS nothing
LIMIT 1
MATCH (orphaned_av:AttributeValue)
WHERE NOT exists((orphaned_av)-[]-())
DELETE orphaned_av
        """ % {"id_func": db.get_id_function_name()}
        self.add_to_query(query)


class DeleteDuplicateBooleanEdgesQuery(Query):
    name = "delete_duplicate_booleans_edges"
    type = QueryType.WRITE
    insert_return = False
    edge_type: DatabaseEdgeType | None = None

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        if not self.edge_type:
            raise RuntimeError("edge_type is required for this query")
        query = """
// -------------------
// find Attribute nodes with multiple identical edges to Boolean nodes
// -------------------
MATCH (a:Attribute)-[e:%(edge_type)s]->(b)
WITH a, e.branch AS branch, e.branch_level AS branch_level, e.status AS status, e.from AS from, e.to AS to, b, COUNT(*) AS num_duplicate_edges
WHERE num_duplicate_edges > 1
// -------------------
// get the identical edges for a given set of Attribute node, edge properties, Boolean
// -------------------
WITH DISTINCT a, branch, branch_level, status, from, to, b
CREATE (a)-[fresh_e:%(edge_type)s {branch: branch, branch_level: branch_level, status: status, from: from}]->(b)
SET fresh_e.to = to
WITH a, branch, status, from, to, b, %(id_func)s(fresh_e) AS e_id_to_keep
CALL {
    WITH a, branch, status, from, to, b, e_id_to_keep
    MATCH (a)-[e:%(edge_type)s]->(b)
    WHERE %(id_func)s(e) <> e_id_to_keep
    AND e.branch = branch AND e.status = status AND e.from = from
    AND (e.to = to OR (e.to IS NULL AND to IS NULL))
    DELETE e
}
        """ % {"edge_type": self.edge_type.value, "id_func": db.get_id_function_name()}
        self.add_to_query(query)


class DeleteDuplicateIsVisibleEdgesQuery(DeleteDuplicateBooleanEdgesQuery):
    name = "delete_duplicate_is_visible_edges"
    type = QueryType.WRITE
    insert_return = False
    edge_type = DatabaseEdgeType.IS_VISIBLE


class DeleteDuplicateIsProtectedEdgesQuery(DeleteDuplicateBooleanEdgesQuery):
    name = "delete_duplicate_is_protected_edges"
    type = QueryType.WRITE
    insert_return = False
    edge_type = DatabaseEdgeType.IS_PROTECTED


class Migration020(GraphMigration):
    """
    1. Find duplicate edges. These can be duplicated if multiple AttributeValue nodes with the same value exist b/c of concurrent
        database updates.
        a. (a:Attribute)-[e:HAS_VALUE]->(av:AttributeValue)
            grouped by (a, e.branch, e.from, e.to, e.status, av.value, av.is_default) to determine the number of duplicates.
        b. (a:Attribute)-[e:HAS_VALUE]->(b:Boolean)
            grouped by (a, e.branch, e.from, e.status, b) to determine the number of duplicates.
    2. For a given set of duplicate edges
        a. delete all of the duplicate edges
        b. merge one edge with the properties of the deleted edges
    3. If there are any orphaned AttributeValue nodes after these changes, then delete them

    This migration does not account for consolidating duplicated AttributeValue nodes because more might be created
    in the future due to concurrent database updates. A migration to consolidate duplicated AttributeValue nodes
    should be run when we find a way to stop duplicate AttributeValue nodes from being created
    """

    name: str = "020_delete_duplicate_edges"
    minimum_version: int = 19
    queries: Sequence[type[Query]] = [
        DeleteDuplicateHasValueEdgesQuery,
        DeleteDuplicateIsVisibleEdgesQuery,
        DeleteDuplicateIsProtectedEdgesQuery,
    ]

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        # skip the transaction b/c it will run out of memory on a large database
        return await self.do_execute(db=db)

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()
        return result



from infrahub.message_bus import InfrahubMessage, InfrahubResponse

from .check_artifact_create import CheckArtifactCreate
from .check_generator_run import CheckGeneratorRun
from .check_repository_checkdefinition import CheckRepositoryCheckDefinition
from .check_repository_mergeconflicts import CheckRepositoryMergeConflicts
from .check_repository_usercheck import CheckRepositoryUserCheck
from .event_branch_create import EventBranchCreate
from .event_branch_delete import EventBranchDelete
from .event_branch_merge import EventBranchMerge
from .event_branch_rebased import EventBranchRebased
from .event_node_mutated import EventNodeMutated
from .event_schema_update import EventSchemaUpdate
from .event_worker_newprimaryapi import EventWorkerNewPrimaryAPI
from .finalize_validator_execution import FinalizeValidatorExecution
from .git_file_get import GitFileGet, GitFileGetResponse
from .git_repository_connectivity import GitRepositoryConnectivity
from .proposed_change.request_proposedchange_refreshartifacts import RequestProposedChangeRefreshArtifacts
from .refresh_git_fetch import RefreshGitFetch
from .refresh_registry_branches import RefreshRegistryBranches
from .refresh_registry_rebasedbranch import RefreshRegistryRebasedBranch
from .request_artifactdefinition_check import RequestArtifactDefinitionCheck
from .request_generatordefinition_check import RequestGeneratorDefinitionCheck
from .request_proposedchange_pipeline import RequestProposedChangePipeline
from .request_repository_checks import RequestRepositoryChecks
from .request_repository_userchecks import RequestRepositoryUserChecks
from .send_echo_request import SendEchoRequest, SendEchoRequestResponse

MESSAGE_MAP: dict[str, type[InfrahubMessage]] = {
    "check.artifact.create": CheckArtifactCreate,
    "check.generator.run": CheckGeneratorRun,
    "check.repository.check_definition": CheckRepositoryCheckDefinition,
    "check.repository.merge_conflicts": CheckRepositoryMergeConflicts,
    "check.repository.user_check": CheckRepositoryUserCheck,
    "event.branch.create": EventBranchCreate,
    "event.branch.delete": EventBranchDelete,
    "event.branch.merge": EventBranchMerge,
    "event.branch.rebased": EventBranchRebased,
    "event.node.mutated": EventNodeMutated,
    "event.schema.update": EventSchemaUpdate,
    "event.worker.new_primary_api": EventWorkerNewPrimaryAPI,
    "finalize.validator.execution": FinalizeValidatorExecution,
    "git.file.get": GitFileGet,
    "git.repository.connectivity": GitRepositoryConnectivity,
    "refresh.git.fetch": RefreshGitFetch,
    "refresh.registry.branches": RefreshRegistryBranches,
    "refresh.registry.rebased_branch": RefreshRegistryRebasedBranch,
    "request.artifact_definition.check": RequestArtifactDefinitionCheck,
    "request.generator_definition.check": RequestGeneratorDefinitionCheck,
    "request.proposed_change.pipeline": RequestProposedChangePipeline,
    "request.proposed_change.refresh_artifacts": RequestProposedChangeRefreshArtifacts,
    "request.repository.checks": RequestRepositoryChecks,
    "request.repository.user_checks": RequestRepositoryUserChecks,
    "send.echo.request": SendEchoRequest,
}

RESPONSE_MAP: dict[str, type[InfrahubResponse]] = {
    "git.file.get": GitFileGetResponse,
    "send.echo.request": SendEchoRequestResponse,
}

PRIORITY_MAP = {
    "check.artifact.create": 2,
    "check.repository.check_definition": 2,
    "check.repository.merge_conflicts": 2,
    "event.branch.create": 5,
    "event.branch.delete": 5,
    "event.branch.merge": 5,
    "event.schema.update": 5,
    "git.diff.names_only": 4,
    "git.file.get": 4,
    "request.artifact.generate": 2,
    "request.git.sync": 4,
    "request.proposed_change.pipeline": 5,
    "transform.jinja.template": 4,
    "transform.python.data": 4,
}


def message_priority(routing_key: str) -> int:
    return PRIORITY_MAP.get(routing_key, 3)


ROUTING_KEY_MAP: dict[type[InfrahubMessage], str] = {
    message: routing_key for routing_key, message in MESSAGE_MAP.items()
}



from typing import TYPE_CHECKING, Any, Optional

from graphene import InputObjectType, Mutation
from graphql import GraphQLResolveInfo
from typing_extensions import Self

from infrahub.core.branch import Branch
from infrahub.core.constants import RESTRICTED_NAMESPACES
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreMenuItem
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import ValidationError
from infrahub.graphql.mutations.main import InfrahubMutationMixin

from .main import InfrahubMutationOptions

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext

EXTENDED_RESTRICTED_NAMESPACES = RESTRICTED_NAMESPACES + ["Builtin"]


def validate_namespace(data: InputObjectType) -> None:
    namespace = data.get("namespace")
    if isinstance(namespace, dict) and "value" in namespace:
        namespace_value = str(namespace.get("value"))
        if namespace_value in EXTENDED_RESTRICTED_NAMESPACES:
            raise ValidationError(
                input_value={"namespace": f"{namespace_value} is not valid, it's a restricted namespace"}
            )


class InfrahubCoreMenuMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(  # pylint: disable=arguments-differ
        cls, schema: NodeSchema, _meta: Optional[Any] = None, **options: dict[str, Any]
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
    ) -> tuple[Node, Self]:
        validate_namespace(data=data)

        obj, result = await super().mutate_create(info=info, data=data, branch=branch)

        return obj, result

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
        node: Optional[Node] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context

        obj = await NodeManager.find_object(
            db=context.db, kind=CoreMenuItem, id=data.get("id"), hfid=data.get("hfid"), branch=branch
        )
        validate_namespace(data=data)

        if obj.protected.value:
            raise ValidationError(input_value="This object is protected, it can't be modified.")

        obj, result = await super().mutate_update(info=info, data=data, branch=branch, node=obj)  # type: ignore[assignment]
        return obj, result  # type: ignore[return-value]

    @classmethod
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context
        obj = await NodeManager.find_object(
            db=context.db, kind=CoreMenuItem, id=data.get("id"), hfid=data.get("hfid"), branch=branch
        )
        if obj.protected.value:
            raise ValidationError(input_value="This object is protected, it can't be deleted.")

        return await super().mutate_delete(info=info, data=data, branch=branch)



from typing import Union

from infrahub_sdk.node import InfrahubNode
from prefect import task
from prefect.cache_policies import NONE

from infrahub import lock
from infrahub.core.constants import InfrahubKind
from infrahub.git.models import RequestArtifactGenerate
from infrahub.message_bus import messages
from infrahub.services import InfrahubServices


@task(name="define-artifact", task_run_name="Define Artifact", cache_policy=NONE)
async def define_artifact(
    message: Union[messages.CheckArtifactCreate, RequestArtifactGenerate], service: InfrahubServices
) -> InfrahubNode:
    if message.artifact_id:
        artifact = await service.client.get(
            kind=InfrahubKind.ARTIFACT, id=message.artifact_id, branch=message.branch_name
        )
    else:
        async with lock.registry.get(f"{message.target_id}-{message.artifact_definition}", namespace="artifact"):
            artifacts = await service.client.filters(
                kind=InfrahubKind.ARTIFACT,
                branch=message.branch_name,
                definition__ids=[message.artifact_definition],
                object__ids=[message.target_id],
            )
            if artifacts:
                artifact = artifacts[0]
            else:
                artifact = await service.client.create(
                    kind=InfrahubKind.ARTIFACT,
                    branch=message.branch_name,
                    data={
                        "name": message.artifact_name,
                        "status": "Pending",
                        "object": message.target_id,
                        "definition": message.artifact_definition,
                        "content_type": message.content_type,
                    },
                )
                await artifact.save()
    return artifact



from infrahub.core.relationship.constraints.profiles_kind import RelationshipProfilesKindConstraint
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class RelationshipProfilesKindConstraintDependency(DependencyBuilder[RelationshipProfilesKindConstraint]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipProfilesKindConstraint:
        return RelationshipProfilesKindConstraint(db=context.db, branch=context.branch)



from typing import Any

from starlette.middleware.cors import CORSMiddleware
from starlette.types import ASGIApp

from infrahub import config


class InfrahubCORSMiddleware(CORSMiddleware):
    def __init__(self, app: ASGIApp, *args: Any, **kwargs: Any) -> None:
        config.SETTINGS.initialize_and_exit()
        kwargs["allow_origins"] = config.SETTINGS.api.cors_allow_origins
        kwargs["allow_credentials"] = config.SETTINGS.api.cors_allow_credentials
        kwargs["allow_methods"] = config.SETTINGS.api.cors_allow_methods
        kwargs["allow_headers"] = config.SETTINGS.api.cors_allow_headers

        super().__init__(app, *args, **kwargs)



from __future__ import annotations

import asyncio
import time
import uuid
from asyncio import Lock as LocalLock
from asyncio import sleep
from typing import TYPE_CHECKING, Optional, Union

import redis.asyncio as redis
from prometheus_client import Histogram
from redis.asyncio.lock import Lock as GlobalLock

from infrahub import config

if TYPE_CHECKING:
    from types import TracebackType

    from infrahub.services import InfrahubServices


registry: InfrahubLockRegistry = None


METRIC_PREFIX = "infrahub_lock"

LOCK_ACQUIRE_TIME_METRICS = Histogram(
    f"{METRIC_PREFIX}_acquire_seconds",
    "Time to acquire the lock on a given object",
    labelnames=["lock", "type"],
    buckets=[0.0005, 0.25, 0.5, 1, 5],
)

LOCK_RESERVE_TIME_METRICS = Histogram(
    f"{METRIC_PREFIX}_reserved_duration_seconds",
    "Time while a given lock is reserved by a given client",
    labelnames=["lock", "type"],
    buckets=[0.001, 0.5, 1, 5, 10],
)

LOCAL_SCHEMA_LOCK = "local.schema"
GLOBAL_INIT_LOCK = "global.init"
GLOBAL_SCHEMA_LOCK = "global.schema"
GLOBAL_GRAPH_LOCK = "global.graph"


class InfrahubMultiLock:
    """Context manager to allow multiple locks to be reserved together"""

    def __init__(self, lock_registry: InfrahubLockRegistry, locks: Optional[list[str]] = None) -> None:
        self.registry = lock_registry
        self.locks = locks or []

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        await self.release()

    async def acquire(self) -> None:
        for lock in self.locks:
            await self.registry.get(name=lock).acquire()

    async def release(self) -> None:
        for lock in reversed(self.locks):
            await self.registry.get(name=lock).release()


class NATSLock:
    """Context manager to lock using NATS"""

    def __init__(self, service: InfrahubServices, name: str) -> None:
        self.name = name
        self.token = None
        self.service = service

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        await self.release()

    async def acquire(self) -> None:
        token = uuid.uuid1().hex
        while True:
            if await self.do_acquire(token):
                self.token = token
                return True
            await sleep(0.1)  # default Redis GlobalLock value

    async def do_acquire(self, token: str) -> bool:
        return await self.service.cache.set(key=self.name, value=token, not_exists=True)

    async def release(self) -> None:
        await self.service.cache.delete(key=self.name)

    async def locked(self) -> bool:
        return await self.service.cache.get(key=self.name) is not None


class InfrahubLock:
    """InfrahubLock object to provide a unified interface for both Local and Distributed locks.

    Having the same interface for both local and distributed tests will simplify our unit tests.
    """

    def __init__(
        self,
        name: str,
        connection: Optional[Union[redis.Redis, InfrahubServices]] = None,
        local: Optional[bool] = None,
        in_multi: bool = False,
    ) -> None:
        self.use_local: bool = local
        self.local: LocalLock = None
        self.remote: GlobalLock = None
        self.name: str = name
        self.connection: Optional[redis.Redis] = connection
        self.in_multi: bool = in_multi
        self.lock_type: str = "multi" if self.in_multi else "individual"
        self.acquire_time: Optional[int] = None
        self.event = asyncio.Event()

        if not self.connection or (self.use_local is None and name.startswith("local.")):
            self.use_local = True

        if self.use_local:
            self.local = LocalLock()
        elif config.SETTINGS.cache.driver == config.CacheDriver.Redis:
            self.remote = GlobalLock(redis=self.connection, name=self.name)
        else:
            self.remote = NATSLock(service=self.connection, name=self.name)

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        await self.release()

    async def acquire(self) -> None:
        with LOCK_ACQUIRE_TIME_METRICS.labels(self.name, self.lock_type).time():
            if not self.use_local:
                await self.remote.acquire()
            else:
                await self.local.acquire()
        self.acquire_time = time.time_ns()
        self.event.clear()

    async def release(self) -> None:
        duration_ns = time.time_ns() - self.acquire_time
        LOCK_RESERVE_TIME_METRICS.labels(self.name, self.lock_type).observe(duration_ns / 1000000000)
        if not self.use_local:
            await self.remote.release()
        else:
            self.local.release()
        self.event.set()

    async def locked(self) -> bool:
        if not self.use_local:
            return await self.remote.locked()

        return self.local.locked()


class InfrahubLockRegistry:
    def __init__(
        self, token: Optional[str] = None, local_only: bool = False, service: Optional[InfrahubServices] = None
    ) -> None:
        if config.SETTINGS.cache.enable and not local_only:
            if config.SETTINGS.cache.driver == config.CacheDriver.Redis:
                self.connection = redis.Redis(
                    host=config.SETTINGS.cache.address,
                    port=config.SETTINGS.cache.service_port,
                    db=config.SETTINGS.cache.database,
                    ssl=config.SETTINGS.cache.tls_enabled,
                    ssl_cert_reqs="optional" if not config.SETTINGS.cache.tls_insecure else "none",
                    ssl_check_hostname=not config.SETTINGS.cache.tls_insecure,
                    ssl_ca_certs=config.SETTINGS.cache.tls_ca_file,
                )
            else:
                self.connection = service
        else:
            self.connection = None

        self.token = token or str(uuid.uuid4())
        self.locks: dict[str, InfrahubLock] = {}

    @classmethod
    def _generate_name(cls, name: str, namespace: Optional[str] = None, local: Optional[bool] = None) -> str:
        if namespace is None and local is None:
            return name

        new_name = ""
        if local is True:
            new_name = "local."
        elif local is False:
            new_name = "global."

        if namespace is not None:
            new_name += f"{namespace}."
        new_name += name

        return new_name

    def get_existing(
        self,
        name: str,
        namespace: str | None,
        local: Optional[bool] = None,
    ) -> InfrahubLock | None:
        lock_name = self._generate_name(name=name, namespace=namespace, local=local)
        if lock_name not in self.locks:
            return None
        return self.locks[lock_name]

    def get(
        self, name: str, namespace: Optional[str] = None, local: Optional[bool] = None, in_multi: bool = False
    ) -> InfrahubLock:
        lock_name = self._generate_name(name=name, namespace=namespace, local=local)
        if lock_name not in self.locks:
            self.locks[lock_name] = InfrahubLock(name=lock_name, connection=self.connection, in_multi=in_multi)
        return self.locks[lock_name]

    def local_schema_lock(self) -> LocalLock:
        return self.get(name=LOCAL_SCHEMA_LOCK)

    def initialization(self) -> LocalLock:
        return self.get(name=GLOBAL_INIT_LOCK)

    async def local_schema_wait(self) -> None:
        await self.get(name=LOCAL_SCHEMA_LOCK).event.wait()

    def global_schema_lock(self) -> InfrahubMultiLock:
        return InfrahubMultiLock(lock_registry=self, locks=[LOCAL_SCHEMA_LOCK, GLOBAL_SCHEMA_LOCK])

    def global_graph_lock(self) -> InfrahubMultiLock:
        return InfrahubMultiLock(lock_registry=self, locks=[LOCAL_SCHEMA_LOCK, GLOBAL_GRAPH_LOCK, GLOBAL_SCHEMA_LOCK])


def initialize_lock(local_only: bool = False, service: Optional[InfrahubServices] = None) -> None:
    global registry  # pylint: disable=global-statement
    registry = InfrahubLockRegistry(local_only=local_only, service=service)


def build_object_lock_name(name: str) -> str:
    return f"global.object.{name}"



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from infrahub.permissions.constants import PermissionDecisionFlag

from . import Node

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class CoreGlobalPermission(Node):
    async def to_graphql(
        self,
        db: InfrahubDatabase,
        fields: Optional[dict] = None,
        related_node_ids: Optional[set] = None,
        filter_sensitive: bool = False,
        permissions: Optional[dict] = None,
        include_properties: bool = True,
    ) -> dict:
        response = await super().to_graphql(
            db,
            fields=fields,
            related_node_ids=related_node_ids,
            filter_sensitive=filter_sensitive,
            permissions=permissions,
            include_properties=include_properties,
        )

        if fields:
            if "identifier" in fields:
                decision = PermissionDecisionFlag(value=self.decision.value.value)  # type: ignore[attr-defined]
                response["identifier"] = {"value": f"global:{self.action.value}:{decision.name.lower()}"}  # type: ignore[attr-defined,union-attr]

        return response


class CoreObjectPermission(Node):
    async def to_graphql(
        self,
        db: InfrahubDatabase,
        fields: Optional[dict] = None,
        related_node_ids: Optional[set] = None,
        filter_sensitive: bool = False,
        permissions: Optional[dict] = None,
        include_properties: bool = True,
    ) -> dict:
        response = await super().to_graphql(
            db,
            fields=fields,
            related_node_ids=related_node_ids,
            filter_sensitive=filter_sensitive,
            permissions=permissions,
            include_properties=include_properties,
        )

        if fields:
            if "identifier" in fields:
                decision = PermissionDecisionFlag(value=self.decision.value.value)  # type: ignore[attr-defined]
                response["identifier"] = {
                    "value": f"object:{self.namespace.value}:{self.name.value}:{self.action.value.value}:{decision.name.lower()}"  # type: ignore[attr-defined,union-attr]
                }

        return response



from __future__ import annotations

from typing import Optional

from ..utils import SubclassWithMeta, SubclassWithMeta_Meta


class BaseOptions:
    name: Optional[str] = None
    description: Optional[str] = None

    _frozen: bool = False

    def __init__(self, class_type):
        self.class_type = class_type

    def freeze(self) -> None:
        self._frozen = True

    def __setattr__(self, name, value):
        if not self._frozen:
            super().__setattr__(name, value)
        else:
            raise Exception(f"Can't modify frozen Options {self}")  # pylint: disable=broad-exception-raised

    def __repr__(self):
        return f"<{self.__class__.__name__} name={repr(self.name)}>"


BaseNodeMeta = SubclassWithMeta_Meta


class BaseNode(SubclassWithMeta):
    # @classmethod
    # def create_node(cls, class_name, **options):
    #     return type(class_name, (cls,), {"Meta": options})

    @classmethod
    def __init_subclass_with_meta__(cls, name=None, description=None, _meta=None, **_kwargs) -> None:
        assert "_meta" not in cls.__dict__, "Can't assign meta directly"
        if not _meta:
            return
        _meta.name = name or cls.__name__
        _meta.description = description
        _meta.freeze()
        cls._meta = _meta
        super().__init_subclass_with_meta__()


class BaseNodeOptions(BaseOptions):
    default_filter = None
    # fields = None  # type: Dict[str, Field]
    # interfaces = ()  # type: Iterable[Type[Interface]]


class ObjectNodeMeta(BaseNodeMeta):
    def __new__(mcs, name_, bases, namespace, **options):  # noqa: N804
        # Note: it's safe to pass options as keyword arguments as they are still type-checked by NodeOptions.

        # We create this type, to then overload it with the dataclass attrs
        class InterObjectNode:
            pass

        base_cls = super().__new__(mcs, name_, (InterObjectNode,) + bases, namespace, **options)
        # if base_cls._meta:
        #     fields = [
        #         (
        #             key,
        #             "typing.Any",
        #             field(
        #                 default=field_value.default_value
        #                 if isinstance(field_value, Field)
        #                 else None
        #             ),
        #         )
        #         for key, field_value in base_cls._meta.fields.items()
        #     ]
        #     dataclass = make_dataclass(name_, fields, bases=())
        #     InterObjectType.__init__ = dataclass.__init__
        #     InterObjectType.__eq__ = dataclass.__eq__
        #     InterObjectType.__repr__ = dataclass.__repr__
        return base_cls



from infrahub import config
from infrahub.auth import AccountSession
from infrahub.core.account import GlobalPermission
from infrahub.core.branch import Branch
from infrahub.core.constants import GlobalPermissions, PermissionDecision
from infrahub.database import InfrahubDatabase
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.initialization import GraphqlParams

from .interface import CheckerResolution, GraphQLQueryPermissionCheckerInterface


class MergeBranchPermissionChecker(GraphQLQueryPermissionCheckerInterface):
    """Checker that makes sure a user account can merge a branch without going through a proposed change."""

    permission_required = GlobalPermission(
        action=GlobalPermissions.MERGE_BRANCH.value, decision=PermissionDecision.ALLOW_ALL.value
    )

    async def supports(self, db: InfrahubDatabase, account_session: AccountSession, branch: Branch) -> bool:
        return config.SETTINGS.main.allow_anonymous_access or account_session.authenticated

    async def check(
        self,
        db: InfrahubDatabase,
        account_session: AccountSession,
        analyzed_query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,
        branch: Branch,
    ) -> CheckerResolution:
        if "BranchMerge" in [operation.name for operation in analyzed_query.operations]:
            query_parameters.context.active_permissions.raise_for_permission(permission=self.permission_required)
            return CheckerResolution.TERMINATE

        return CheckerResolution.NEXT_CHECKER



from typing import Optional

from pydantic import Field

from infrahub.generators.models import ProposedChangeGeneratorDefinition
from infrahub.message_bus import InfrahubMessage


class CheckGeneratorRun(InfrahubMessage):
    """A check that runs a generator."""

    generator_definition: ProposedChangeGeneratorDefinition = Field(..., description="The Generator definition")
    generator_instance: Optional[str] = Field(
        default=None, description="The id of the generator instance if it previously existed"
    )
    commit: str = Field(..., description="The commit to target")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    repository_kind: str = Field(..., description="The kind of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    target_id: str = Field(..., description="The ID of the target object for this generator")
    target_name: str = Field(..., description="Name of the generator target")
    query: str = Field(..., description="The name of the query to use when collecting data")
    variables: dict = Field(..., description="Input variables when running the generator")
    validator_id: str = Field(..., description="The ID of the validator")
    proposed_change: str | None = Field(None, description="The unique ID of the Proposed Change")



from infrahub.core.diff.parent_node_adder import DiffParentNodeAdder
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffParentNodeAdderDependency(DependencyBuilder[DiffParentNodeAdder]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffParentNodeAdder:
        return DiffParentNodeAdder()



from dataclasses import dataclass
from typing import Optional

from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import RelationshipCardinality, RelationshipDirection
from infrahub.core.query.relationship import RelationshipCountPerNodeQuery
from infrahub.core.schema import MainSchemaTypes
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import ValidationError

from ..model import RelationshipManager
from .interface import RelationshipManagerConstraintInterface


@dataclass
class NodeToValidate:
    uuid: str
    cardinality: RelationshipCardinality
    min_count: Optional[int] = None
    max_count: Optional[int] = None


class RelationshipCountConstraint(RelationshipManagerConstraintInterface):
    def __init__(self, db: InfrahubDatabase, branch: Optional[Branch] = None):
        self.db = db
        self.branch = branch

    async def check(self, relm: RelationshipManager, node_schema: MainSchemaTypes) -> None:
        branch = await registry.get_branch(db=self.db) if not self.branch else self.branch

        # NOTE adding resolve here because we need to retrieve the real ID
        # but if the validation fails we'll end up with some allocated resources that are not being used
        await relm.resolve(db=self.db)

        nodes_to_validate: list[NodeToValidate] = []

        # peer_ids_present_local_only:
        #    new relationship, need to check if the schema on the other side has a max_count defined
        # peer_ids_present_database_only:
        #    relationship to be deleted, need to check if the schema on the other side has a min_count defined
        # TODO see how to manage Generic node
        peer_schema = registry.schema.get(name=relm.schema.peer, branch=branch)
        peer_rels = peer_schema.get_relationships_by_identifier(id=relm.schema.get_identifier())
        if not peer_rels:
            return

        update_details = await relm.fetch_relationship_ids(db=self.db, force_refresh=False)
        for peer_rel in peer_rels:
            # If a relationship is directional and both have the same direction they can't work together
            if relm.schema.direction == peer_rel.direction and peer_rel.direction != RelationshipDirection.BIDIR:
                continue

            for peer_id in update_details.peer_ids_present_local_only + update_details.peer_ids_present_database_only:
                if peer_rel.max_count and peer_id in update_details.peer_ids_present_local_only:
                    nodes_to_validate.append(
                        NodeToValidate(uuid=peer_id, max_count=peer_rel.max_count, cardinality=peer_rel.cardinality)
                    )

                if peer_rel.min_count and peer_id in update_details.peer_ids_present_database_only:
                    nodes_to_validate.append(
                        NodeToValidate(uuid=peer_id, min_count=peer_rel.min_count, cardinality=peer_rel.cardinality)
                    )

        query = await RelationshipCountPerNodeQuery.init(
            db=self.db,
            node_ids=[node.uuid for node in nodes_to_validate],
            identifier=relm.schema.identifier,
            direction=relm.schema.direction.neighbor_direction,
            branch=branch,
        )
        await query.execute(db=self.db)
        count_per_peer = await query.get_count_per_peer()

        # Need to adjust the number based on what we will add / remove
        #  +1 for max_count
        #  -1 for min_count
        for node in nodes_to_validate:
            if node.max_count and count_per_peer[node.uuid] + 1 > node.max_count:
                raise ValidationError(
                    f"Node {node.uuid} has {count_per_peer[node.uuid] + 1} peers "
                    f"for {relm.schema.identifier}, maximum of {node.max_count} allowed",
                )
            if node.min_count and count_per_peer[node.uuid] - 1 < node.min_count:
                raise ValidationError(
                    f"Node {node.uuid} has {count_per_peer[node.uuid] - 1} peers "
                    f"for {relm.schema.identifier}, no fewer than {node.min_count} allowed",
                )



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class RequestRepositoryChecks(InfrahubMessage):
    """Sent to trigger the checks for a repository to be executed."""

    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    repository: str = Field(..., description="The unique ID of the Repository")
    source_branch: str = Field(..., description="The source branch")
    target_branch: str = Field(..., description="The target branch")



from uuid import uuid4

from infrahub.core.constants import NULL_VALUE, DiffAction, RelationshipCardinality
from infrahub.core.constants.database import DatabaseEdgeType

from .model.path import (
    EnrichedDiffAttribute,
    EnrichedDiffConflict,
    EnrichedDiffNode,
    EnrichedDiffProperty,
    EnrichedDiffRelationship,
    EnrichedDiffRoot,
    EnrichedDiffSingleRelationship,
)


class ConflictsEnricher:
    def __init__(self) -> None:
        self._base_branch_name: str | None = None
        self._diff_branch_name: str | None = None

    @property
    def base_branch_name(self) -> str:
        if not self._base_branch_name:
            raise RuntimeError("could not identify base branch")
        return self._base_branch_name

    @property
    def diff_branch_name(self) -> str:
        if not self._diff_branch_name:
            raise RuntimeError("could not identify diff branch")
        return self._diff_branch_name

    async def add_conflicts_to_branch_diff(
        self, base_diff_root: EnrichedDiffRoot, branch_diff_root: EnrichedDiffRoot
    ) -> None:
        self._base_branch_name = branch_diff_root.base_branch_name
        self._diff_branch_name = branch_diff_root.diff_branch_name

        base_node_map = {n.uuid: n for n in base_diff_root.nodes}
        branch_node_map = {n.uuid: n for n in branch_diff_root.nodes}
        branch_node_map_uuids = set(branch_node_map.keys())
        common_node_uuids = set(base_node_map.keys()) & branch_node_map_uuids
        for node_uuid in common_node_uuids:
            base_node = base_node_map[node_uuid]
            branch_node = branch_node_map[node_uuid]
            self._add_node_conflicts(base_node=base_node, branch_node=branch_node)
        # remove conflicts from branch nodes that have been manually corrected on the base branch
        for branch_only_node_uuid in branch_node_map_uuids - common_node_uuids:
            branch_node = branch_node_map[branch_only_node_uuid]
            branch_node.clear_conflicts()

    def _add_node_conflicts(self, base_node: EnrichedDiffNode, branch_node: EnrichedDiffNode) -> None:
        if base_node.action != branch_node.action and DiffAction.UNCHANGED not in {
            base_node.action,
            branch_node.action,
        }:
            self._add_node_conflict(base_node=base_node, branch_node=branch_node)
        elif branch_node.conflict:
            branch_node.conflict = None
        # adding attr/rel conflicts when there is an unresolvable node-level conflict is pointless
        if branch_node.conflict and branch_node.conflict.resolvable is False:
            return
        base_attribute_map = {a.name: a for a in base_node.attributes}
        branch_attribute_map = {a.name: a for a in branch_node.attributes}
        common_attribute_names = set(base_attribute_map.keys()) & set(branch_attribute_map.keys())
        for branch_attribute in branch_node.attributes:
            if branch_attribute.name in common_attribute_names:
                base_attribute = base_attribute_map[branch_attribute.name]
                self._add_attribute_conflicts(base_attribute=base_attribute, branch_attribute=branch_attribute)
            else:
                branch_attribute.clear_conflicts()
        base_relationship_map = {r.name: r for r in base_node.relationships}
        branch_relationship_map = {r.name: r for r in branch_node.relationships}
        common_relationship_names = set(base_relationship_map.keys()) & set(branch_relationship_map.keys())
        for branch_relationship in branch_node.relationships:
            if branch_relationship.name in common_relationship_names:
                base_relationship = base_relationship_map[branch_relationship.name]
                self._add_relationship_conflicts(
                    base_relationship=base_relationship,
                    branch_relationship=branch_relationship,
                )
            else:
                branch_relationship.clear_conflicts()

    def _add_node_conflict(self, base_node: EnrichedDiffNode, branch_node: EnrichedDiffNode) -> None:
        if branch_node.conflict:
            conflict_uuid = branch_node.conflict.uuid
            selected_branch = branch_node.conflict.selected_branch
        else:
            conflict_uuid = str(uuid4())
            selected_branch = None
        resolvable = True
        # this condition should always be true, but it's good to be explicit
        if DiffAction.REMOVED in [base_node.action, branch_node.action]:
            resolvable = False
        branch_node.conflict = EnrichedDiffConflict(
            uuid=conflict_uuid,
            base_branch_action=base_node.action,
            base_branch_value=None,
            base_branch_changed_at=base_node.changed_at,
            diff_branch_action=branch_node.action,
            diff_branch_value=None,
            diff_branch_changed_at=branch_node.changed_at,
            selected_branch=selected_branch,
            resolvable=resolvable,
        )

    def _add_attribute_conflicts(
        self,
        base_attribute: EnrichedDiffAttribute,
        branch_attribute: EnrichedDiffAttribute,
    ) -> None:
        base_property_map = {p.property_type: p for p in base_attribute.properties}
        branch_property_map = {p.property_type: p for p in branch_attribute.properties}
        common_property_types = set(base_property_map.keys()) & set(branch_property_map.keys())
        for branch_property in branch_attribute.properties:
            if branch_property.property_type not in common_property_types:
                branch_property.conflict = None
                continue
            base_property = base_property_map[branch_property.property_type]
            property_actions = {branch_property.action, base_property.action}
            if DiffAction.UNCHANGED in property_actions:
                branch_property.conflict = None
                continue
            same_value = self._have_same_value(base_property=base_property, branch_property=branch_property)
            if same_value:
                branch_property.conflict = None
                continue
            self._add_property_conflict(
                base_property=base_property,
                branch_property=branch_property,
            )

    def _add_relationship_conflicts(
        self,
        base_relationship: EnrichedDiffRelationship,
        branch_relationship: EnrichedDiffRelationship,
    ) -> None:
        is_cardinality_one = branch_relationship.cardinality is RelationshipCardinality.ONE
        if is_cardinality_one:
            if not base_relationship.relationships or not branch_relationship.relationships:
                branch_relationship.clear_conflicts()
                return
            base_element = next(iter(base_relationship.relationships))
            branch_element = next(iter(branch_relationship.relationships))
            self._add_relationship_conflicts_for_one_peer(
                base_element=base_element,
                branch_element=branch_element,
                is_cardinality_one=is_cardinality_one,
            )
            return
        base_peer_id_map = {element.peer_id: element for element in base_relationship.relationships}
        branch_peer_id_map = {element.peer_id: element for element in branch_relationship.relationships}
        common_peer_ids = set(base_peer_id_map.keys()) & set(branch_peer_id_map.keys())
        for branch_element in branch_relationship.relationships:
            if branch_element.peer_id not in common_peer_ids:
                branch_element.clear_conflicts()
                continue
            base_element = base_peer_id_map[branch_element.peer_id]
            self._add_relationship_conflicts_for_one_peer(
                base_element=base_element,
                branch_element=branch_element,
                is_cardinality_one=is_cardinality_one,
            )

    def _add_relationship_conflicts_for_one_peer(
        self,
        base_element: EnrichedDiffSingleRelationship,
        branch_element: EnrichedDiffSingleRelationship,
        is_cardinality_one: bool,
    ) -> None:
        base_properties_by_type = {p.property_type: p for p in base_element.properties}
        branch_properties_by_type = {p.property_type: p for p in branch_element.properties}
        common_property_types = set(base_properties_by_type.keys()) & set(branch_properties_by_type.keys())
        for branch_property in branch_element.properties:
            if branch_property.property_type not in common_property_types:
                branch_property.conflict = None
                continue
            base_property = base_properties_by_type[branch_property.property_type]
            includes_unchanged = DiffAction.UNCHANGED in {branch_property.action, base_property.action}
            same_value = self._have_same_value(base_property=base_property, branch_property=branch_property)
            # special handling for cardinality-one peer ID conflict
            if branch_property.property_type is DatabaseEdgeType.IS_RELATED and is_cardinality_one:
                if same_value or includes_unchanged:
                    branch_element.conflict = None
                    branch_property.conflict = None
                    continue
                if branch_element.conflict:
                    conflict_uuid = branch_element.conflict.uuid
                    selected_branch = branch_element.conflict.selected_branch
                else:
                    conflict_uuid = str(uuid4())
                    selected_branch = None
                conflict = EnrichedDiffConflict(
                    uuid=conflict_uuid,
                    base_branch_action=base_element.action,
                    base_branch_value=base_property.new_value,
                    base_branch_changed_at=base_property.changed_at,
                    diff_branch_action=branch_element.action,
                    diff_branch_value=branch_property.new_value,
                    diff_branch_changed_at=branch_property.changed_at,
                    selected_branch=selected_branch,
                )
                branch_element.conflict = conflict
                continue
            if same_value or includes_unchanged:
                branch_property.conflict = None
                continue
            if branch_property.conflict:
                conflict_uuid = branch_property.conflict.uuid
                selected_branch = branch_property.conflict.selected_branch
            else:
                conflict_uuid = str(uuid4())
                selected_branch = None
            branch_property.conflict = EnrichedDiffConflict(
                uuid=conflict_uuid,
                base_branch_action=base_property.action,
                base_branch_value=base_property.new_value,
                base_branch_changed_at=base_property.changed_at,
                diff_branch_action=branch_property.action,
                diff_branch_value=branch_property.new_value,
                diff_branch_changed_at=branch_property.changed_at,
                selected_branch=selected_branch,
            )

    def _add_property_conflict(
        self,
        base_property: EnrichedDiffProperty,
        branch_property: EnrichedDiffProperty,
    ) -> None:
        if branch_property.conflict:
            conflict_uuid = branch_property.conflict.uuid
            selected_branch = branch_property.conflict.selected_branch
        else:
            conflict_uuid = str(uuid4())
            selected_branch = None
        branch_property.conflict = EnrichedDiffConflict(
            uuid=conflict_uuid,
            base_branch_action=base_property.action,
            base_branch_value=base_property.new_value,
            base_branch_changed_at=base_property.changed_at,
            diff_branch_action=branch_property.action,
            diff_branch_value=branch_property.new_value,
            diff_branch_changed_at=branch_property.changed_at,
            selected_branch=selected_branch,
        )

    def _have_same_value(self, base_property: EnrichedDiffProperty, branch_property: EnrichedDiffProperty) -> bool:
        if base_property.new_value == branch_property.new_value:
            return True
        if {base_property.new_value, branch_property.new_value} <= {NULL_VALUE, None}:
            return True
        if (
            base_property.action is DiffAction.UNCHANGED
            and base_property.previous_value == branch_property.previous_value
        ):
            return True
        return False



from __future__ import annotations

from infrahub.exceptions import ValidationError
from infrahub.utils import InfrahubStringEnum


class ProposedChangeState(InfrahubStringEnum):
    OPEN = "open"
    MERGED = "merged"
    MERGING = "merging"
    CLOSED = "closed"
    CANCELED = "canceled"

    def validate_state_check_run(self) -> None:
        if self == ProposedChangeState.OPEN:
            return

        raise ValidationError(input_value="Unable to trigger check on proposed changes that aren't in the open state")

    def validate_editability(self) -> None:
        if self in [ProposedChangeState.CANCELED, ProposedChangeState.MERGED, ProposedChangeState.MERGED]:
            raise ValidationError(
                input_value=f"A proposed change in the {self.value} state is not allowed to be updated"
            )

    def validate_state_transition(self, updated_state: ProposedChangeState) -> None:
        if self == ProposedChangeState.OPEN:
            return

        if self == ProposedChangeState.CLOSED and updated_state not in [
            ProposedChangeState.CANCELED,
            ProposedChangeState.OPEN,
        ]:
            raise ValidationError(
                input_value="A closed proposed change is only allowed to transition to the open state"
            )



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Optional, TypeVar

import ujson
from infrahub_sdk.uuidt import UUIDT

from infrahub.dependencies.registry import build_component_registry
from infrahub.message_bus import InfrahubMessage, Meta
from infrahub.message_bus.messages import ROUTING_KEY_MAP
from infrahub.message_bus.operations import execute_message
from infrahub.services import InfrahubServices
from infrahub.services.adapters.message_bus import InfrahubMessageBus

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase
    from infrahub.message_bus.types import MessageTTL

ResponseClass = TypeVar("ResponseClass")


class BusSimulator(InfrahubMessageBus):
    def __init__(self, database: Optional[InfrahubDatabase] = None) -> None:
        self.messages: list[InfrahubMessage] = []
        self.messages_per_routing_key: dict[str, list[InfrahubMessage]] = {}
        self.service: InfrahubServices = InfrahubServices(database=database, message_bus=self)
        self.replies: dict[str, list[InfrahubMessage]] = defaultdict(list)
        build_component_registry()

    async def publish(
        self, message: InfrahubMessage, routing_key: str, delay: Optional[MessageTTL] = None, is_retry: bool = False
    ) -> None:
        self.messages.append(message)
        if routing_key not in self.messages_per_routing_key:
            self.messages_per_routing_key[routing_key] = []
        self.messages_per_routing_key[routing_key].append(message)
        await execute_message(routing_key=routing_key, message_body=message.body, service=self.service)

    async def reply(self, message: InfrahubMessage, routing_key: str) -> None:
        correlation_id = message.meta.correlation_id or "default"
        self.replies[correlation_id].append(message)

    async def rpc(self, message: InfrahubMessage, response_class: type[ResponseClass]) -> ResponseClass:
        routing_key = ROUTING_KEY_MAP.get(type(message), "")

        correlation_id = str(UUIDT())
        message.meta = Meta(correlation_id=correlation_id, reply_to="ci-testing")

        await self.publish(message=message, routing_key=routing_key)
        reply_id = correlation_id or "default"
        assert len(self.replies[reply_id]) == 1
        response = self.replies[reply_id][0]
        data = ujson.loads(response.body)
        return response_class(**data)

    @property
    def seen_routing_keys(self) -> list[str]:
        return list(self.messages_per_routing_key.keys())



from pydantic import Field

from infrahub.message_bus import InfrahubMessage, InfrahubResponse, InfrahubResponseData

ROUTING_KEY = "send.echo.request"


class SendEchoRequest(InfrahubMessage):
    """Sent a echo request, i.e., ping message."""

    message: str = Field(..., description="The message to send")


class SendEchoRequestResponseData(InfrahubResponseData):
    response: str = Field(default="", description="The response string")


class SendEchoRequestResponse(InfrahubResponse):
    routing_key: str = ROUTING_KEY
    data: SendEchoRequestResponseData



from infrahub.core.diff.ipam_diff_parser import IpamDiffParser
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from ..ip.kinds_getter import IpamKindsGetterDependency
from .repository import DiffRepositoryDependency


class IpamDiffParserDependency(DependencyBuilder[IpamDiffParser]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> IpamDiffParser:
        return IpamDiffParser(
            db=context.db,
            diff_repository=DiffRepositoryDependency.build(context=context),
            ip_kinds_getter=IpamKindsGetterDependency.build(context=context),
        )



from typing import Any

from pydantic import Field

from infrahub.core.constants import MutationAction
from infrahub.message_bus import InfrahubMessage

from .models import InfrahubBranchEvent


class NodeMutatedEvent(InfrahubBranchEvent):
    """Event generated when a node has been mutated"""

    kind: str = Field(..., description="The type of object modified")
    node_id: str = Field(..., description="The ID of the mutated node")
    action: MutationAction = Field(..., description="The action taken on the node")
    data: dict[str, Any] = Field(..., description="Data on modified object")
    fields: list[str] = Field(default_factory=list, description="Fields provided in the mutation")

    def get_name(self) -> str:
        return f"{self.get_event_namespace()}.node.{self.action.value}"

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.node.{self.node_id}",
            "infrahub.node.kind": self.kind,
            "infrahub.node.id": self.node_id,
            "infrahub.node.action": self.action.value,
            "infrahub.branch.name": self.branch,
        }

    def get_payload(self) -> dict[str, Any]:
        return {"data": self.data, "fields": self.fields}

    def get_messages(self) -> list[InfrahubMessage]:
        return [
            # EventNodeMutated(
            #     branch=self.branch,
            #     kind=self.kind,
            #     node_id=self.node_id,
            #     action=self.action.value,
            #     data=self.data,
            #     meta=self.get_message_meta(),
            # )
        ]



from __future__ import annotations

import asyncio
import importlib
import os
import warnings
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

import ujson
from pydantic import BaseModel, Field

from infrahub_sdk.repository import GitRepoManager

from .exceptions import UninitializedError

if TYPE_CHECKING:
    from . import InfrahubClient

INFRAHUB_CHECK_VARIABLE_TO_IMPORT = "INFRAHUB_CHECKS"

_client_class = "InfrahubClient"


class InfrahubCheckInitializer(BaseModel):
    """Information about the originator of the check."""

    proposed_change_id: str = Field(
        default="", description="If available the ID of the proposed change that requested the check"
    )


class InfrahubCheck:
    name: str | None = None
    query: str = ""
    timeout: int = 10

    def __init__(
        self,
        branch: str | None = None,
        root_directory: str = "",
        output: str | None = None,
        initializer: InfrahubCheckInitializer | None = None,
        params: dict | None = None,
        client: InfrahubClient | None = None,
    ):
        self.git: GitRepoManager | None = None
        self.initializer = initializer or InfrahubCheckInitializer()

        self.logs: list[dict[str, Any]] = []
        self.passed = False

        self.output = output

        self.branch = branch
        self.params = params or {}

        self.root_directory = root_directory or os.getcwd()

        self._client = client

        if not self.name:
            self.name = self.__class__.__name__

        if not self.query:
            raise ValueError("A query must be provided")

    def __str__(self) -> str:
        return self.__class__.__name__

    @property
    def client(self) -> InfrahubClient:
        if self._client:
            return self._client

        raise UninitializedError(message="This check has not been initialized with a client")

    @client.setter
    def client(self, value: InfrahubClient) -> None:
        self._client = value

    @classmethod
    async def init(cls, client: InfrahubClient | None = None, *args: Any, **kwargs: Any) -> InfrahubCheck:
        """Async init method, If an existing InfrahubClient client hasn't been provided, one will be created automatically."""
        warnings.warn(
            "InfrahubCheck.init has been deprecated and will be removed in the version in Infrahub SDK 2.0.0",
            DeprecationWarning,
            stacklevel=1,
        )
        if not client:
            client_module = importlib.import_module("infrahub_sdk.client")
            client_class = getattr(client_module, _client_class)
            client = client_class()
        kwargs["client"] = client
        return cls(*args, **kwargs)

    @property
    def errors(self) -> list[dict[str, Any]]:
        return [log for log in self.logs if log["level"] == "ERROR"]

    def _write_log_entry(
        self, message: str, level: str, object_id: str | None = None, object_type: str | None = None
    ) -> None:
        log_message = {"level": level, "message": message, "branch": self.branch_name}
        if object_id:
            log_message["object_id"] = object_id
        if object_type:
            log_message["object_type"] = object_type
        self.logs.append(log_message)

        if self.output == "stdout":
            print(ujson.dumps(log_message))

    def log_error(self, message: str, object_id: str | None = None, object_type: str | None = None) -> None:
        self._write_log_entry(message=message, level="ERROR", object_id=object_id, object_type=object_type)

    def log_info(self, message: str, object_id: str | None = None, object_type: str | None = None) -> None:
        self._write_log_entry(message=message, level="INFO", object_id=object_id, object_type=object_type)

    @property
    def log_entries(self) -> str:
        output = ""
        for log in self.logs:
            output += "-----------------------\n"
            output += f"Message: {log['message']}\n"
            output += f"Level: {log['level']}\n"
            if "object_id" in log:
                output += f"Object ID: {log['object_id']}\n"
            if "object_type" in log:
                output += f"Object ID: {log['object_type']}\n"
        return output

    @property
    def branch_name(self) -> str:
        """Return the name of the current git branch."""

        if self.branch:
            return self.branch

        if not self.git:
            self.git = GitRepoManager(self.root_directory)

        self.branch = str(self.git.active_branch)
        return self.branch

    @abstractmethod
    def validate(self, data: dict) -> None:
        """Code to validate the status of this check."""

    async def collect_data(self) -> dict:
        """Query the result of the GraphQL Query defined in self.query and return the result"""

        return await self.client.query_gql_query(name=self.query, branch_name=self.branch_name, variables=self.params)

    async def run(self, data: dict | None = None) -> bool:
        """Execute the check after collecting the data from the GraphQL query.
        The result of the check is determined based on the presence or not of ERROR log messages."""

        if not data:
            data = await self.collect_data()
        unpacked = data.get("data") or data

        if asyncio.iscoroutinefunction(self.validate):
            await self.validate(data=unpacked)
        else:
            self.validate(data=unpacked)

        nbr_errors = len([log for log in self.logs if log["level"] == "ERROR"])

        self.passed = bool(nbr_errors == 0)

        if self.passed:
            self.log_info("Check succesfully completed")

        return self.passed



from asyncio import run as aiorun

import typer

from infrahub import config
from infrahub.cli.context import CliContext
from infrahub.cli.db import app as db_app
from infrahub.cli.events import app as events_app
from infrahub.cli.git_agent import app as git_app
from infrahub.cli.server import app as server_app
from infrahub.cli.tasks import app as tasks_app
from infrahub.core.initialization import initialization
from infrahub.database import InfrahubDatabase, get_db

# pylint: disable=import-outside-toplevel

app = typer.Typer(name="Infrahub CLI", pretty_exceptions_enable=False)


@app.callback()
def common(ctx: typer.Context) -> None:
    """Infrahub CLI"""
    ctx.obj = CliContext()


app.add_typer(server_app, name="server")
app.add_typer(git_app, name="git-agent")
app.add_typer(db_app, name="db")
app.add_typer(events_app, name="events", help="Interact with the events system.")
app.add_typer(tasks_app, name="tasks", hidden=True)


async def _init_shell(config_file: str) -> None:
    """Launch a Python Interactive shell."""
    config.load_and_exit(config_file_name=config_file)

    db = InfrahubDatabase(driver=await get_db(retry=1))

    async with db.start_session() as db:
        await initialization(db=db)


@app.command()
def shell(config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG")) -> None:
    """Start a python shell within Infrahub context."""
    aiorun(_init_shell(config_file=config_file))

    # TODO add check to properly exit of ipython is not installed
    from IPython import embed
    from rich import pretty
    from traitlets.config import get_config

    pretty.install()

    c = get_config()
    c.InteractiveShellEmbed.colors = "Linux"
    c.InteractiveShellApp.extensions.append("rich")
    embed(config=c)
